#netzMadeDifficult

import numpy as n
import math as math
import random as random
import matplotlib.pyplot as plt

class Netz:
    # Konstruktor mit den am meisten im Test benutzten Werten als Default
    def __init__(self, rad = 20, num = 3, xcor=1, ycor=1, NetzTyp = 0):

        # Koordinaten des Netzmittelpunktes
        self.Ursprung = [float(xcor), float(ycor)]
        # ungefährer Radius des Netzes
        self.Radius = rad
        # kürzeste Speiche, wichtig zum berechnen der (nicht-Netzfüllenden) Spirale
        # wird zu Beginn (als Aproximation) gleich "rad" gesetzt
        self.short = rad

        if (NetzTyp == 0):
            self.makeOutside(num)
            self.makeInside() 
            self.makeSpeichen()
            self.makeFlexSpiral(32)

        if (NetzTyp == 1):
            self.makeOutside(num)
            self.makeInside() 
            self.makeSpeichen()
            self.makeRoundSpiral(32)

        #plt.show()
        

    # Berechnet den Winkel zwischen (xcor|ycor), dem Netzmittelpunkt und dem Horizont durch den Netzmittelpunkt
    def getAngle(self, xcor, ycor):
        #Koordinaten relativ zum netzmittelpunkt
        xkor = float(xcor) - self.Ursprung[0]
        ykor = float(ycor) - self.Ursprung[1]
        #Sinnvoller Error
        if (xkor == 0 and ykor == 0):
            raise Exception("Error, Origin has no angle to itself (returned 0, maybe)")
        
        #Alternative, falls den Code abzubrechen nicht gut ist
        #if (xkor == 0 and ykor == 0):
        #    print("Error, Origin has no angle to itself (returned 0)")
        #    return 0
        
        #Winkel <= Pi
        #erkennbar daran, dass der Punkt über dem Netzmittelpunkt liegt
        if(ykor >= 0):
            return (math.acos(xkor / pow((xkor * xkor + ykor * ykor), 0.5)))
        #Winkel > Pi 
        else:   
            return (2 * math.pi - math.acos(xkor / pow((xkor * xkor + ykor * ykor), 0.5)))

    # berechnet plottet die Grundstruktur 
    def makeOutside(self, num):

        # x und y koordinaten der Grundstruktur / des äußeren Rings des Netzes
        self.xNetz = []
        self.yNetz = []

        # berechnet gleichmäßig auf einem Kreis liegende punkte, die dann in einem geringen maß zufällig gestreut werden
        # Zufallseinfluss nimmt zu, je weniger Außen-Eckpunkte es gibt ("num" im Konstruktor)
        for i in range (num):
            angle = math.pi / num * i * 2
            self.xNetz.append(self.Ursprung[0] + (self.Radius * math.cos(angle)))
            self.yNetz.append(self.Ursprung[1] + (self.Radius * math.sin(angle)))
            aengl = random.random() * 2 * math.pi
            raeng = random.random() * self.Radius / num * 2
            self.xNetz[i] += raeng * math.cos(aengl)
            self.yNetz[i] += raeng * math.sin(aengl)

        plt.plot([*self.xNetz, self.xNetz[0]], [*self.yNetz, self.yNetz[0]], linewidth = 0.4, color = 'k')

        self.Ursprung[0] = Average(self.xNetz)
        self.Ursprung[1] = Average(self.yNetz)

    # berechnet und plottet die Eckpunkte des Inneren Netz"rings" /
    # schneidet die Ecken des äußeren Rings von der Spirale und den Speiche ab
    def makeInside(self):

        # mehr listen von Koordinaten
        self.xInside = []
        self.yInside = []

        for i in range (len(self.yNetz)):

            # Modulo verhindert, das der übergang vom Ende zurück zum Anfang ein Problem wird
            k = (i + 1 ) % len(self.yNetz)

            # die Ecken werden bei 1/3 und 2/3 der Strecke abgeschnitten
            self.xInside.append((self.xNetz[i] + self.xNetz[i] + self.xNetz[k]) / 3)
            self.yInside.append((self.yNetz[i] + self.yNetz[i] + self.yNetz[k]) / 3)

            self.xInside.append((self.xNetz[i] + self.xNetz[k] + self.xNetz[k]) / 3)
            self.yInside.append((self.yNetz[i] + self.yNetz[k] + self.yNetz[k]) / 3)

        plt.plot([*self.xInside, self.xInside[0]], [*self.yInside, self.yInside[0]], linewidth = 0.4, color = 'k')


    # berechnet die Fußpunkte der Speichen auf den inneren Ring und plottet die Speichen
    def makeSpeichen(self):
        
        # self.Angles speichert die Winkel in denen die Speichen vom Mittelpunkt abgehen
        self.Angles = self.makeAngle(27)
        # noch mehr Listen von Koordinaten
        self.xSpeichen = []
        self.ySpeichen = []

        # berechnet die Fußpunkte:
        # geht alle Speichen(-winkel) durch
        for i in range (len(self.Angles)):
            side = len(self.xInside) - 1

            # Sucht die Seite des inneren Rings, auf dem der Fußpunkt liegt
            # in Form der Eckpunkte, die die Seite begrenzen
            for j in range (len(self.xInside)):
                
                # bei j = len: k = 0
                k = (j + 1) % len(self.xInside)
                aenglleJ = self.getAngle(self.xInside[j], self.yInside[j])
                aenglleK = self.getAngle(self.xInside[k], self.yInside[k])

                # wenn der Winkel der Speiche genau zwischen denen der Ecken liegt...
                #### f e c k  t h i s
                if ((0 < self.Angles[i] - aenglleJ) and (aenglleK - aenglleJ > self.Angles[i] - aenglleJ)):

                    # haben wir die richtige Seite gefunden
                    side = j
                    break
            
            # I liegt auf der Strecke vom Eckpunkt "side" zum Eckpunkt "side + 1"
            # damit haben wir 6 von 8 Koordinaten : (self.Ursprung[0]|self.Ursprung[1]), (self.xInside[side]|self.yInside[side]) und 
            # (self.xInside[side+1]|self.yInside[side+1]) (nicht unbedingt in der Reihenfolge)
            # Fehlen nur noch zwei:
            # die des (eines) zweiten Punktes auf der Speiche
            lastx = self.Ursprung[0] + math.cos(self.Angles[i])
            lasty = self.Ursprung[1] + math.sin(self.Angles[i])

            # komisches Tupel, weil line_Intersection nun mal ein Tupel zurückgibt...
            SchnittTupel = line_intersection(self.Ursprung[0], lastx, self.xInside[side], self.xInside[(side + 1) % len(self.xInside)], self.Ursprung[1], lasty, self.yInside[side], self.yInside[(side + 1) % len(self.yInside)])
            # das dann ausgelesen werden muss
            self.xSpeichen.append(SchnittTupel[0])
            self.ySpeichen.append(SchnittTupel[1])

            # findet die kuerzeste Speiche
            if (Pythagoras((self.xSpeichen[i] - self.Ursprung[0]), (self.Ursprung[1] - self.ySpeichen[i])) < self.short):
                self.short = Pythagoras((self.xSpeichen[i] - self.Ursprung[0]), (self.Ursprung[1] - self.ySpeichen[i]))
                print(self.short)

            plt.plot((self.Ursprung[0], self.xSpeichen[i]), (self.Ursprung[1], self.ySpeichen[i]), linewidth = 0.4, color = 'k')
    
    # erzeugt eine liste von zufälligen Winkeln für die Speichen               
    def makeAngle(self, num):
        retAngle = []
        #wir recyclen den _init_, um zufällige, aber nicht *zu* zufällige Winkel zu erhalten.
        for i in range (num):
            angle = math.pi * 2 * i / num 
            aengl = random.random() * 2 * math.pi
            raeng = random.random() / num

            xtemp = (self.Ursprung[0] + (1 * math.cos(angle)) + (raeng * math.cos(aengl)))
            ytemp = (self.Ursprung[1] + (1 * math.sin(angle)) + (raeng * math.sin(aengl)))

            retAngle.append(self.getAngle(xtemp, ytemp))
            
        return retAngle         
    
    # berechnet und plottet die Fangspirale
    def makeFlexSpiral(self, windung):
        self.xSpiral = []
        self.ySpiral = []
        ragne = len(self.xSpeichen) * windung
        for i in range (ragne):
            k = i % len(self.xSpeichen)
            position = ((i + 20) / (ragne + 40))
            self.xSpiral.append(self.Ursprung[0] + ((self.xSpeichen[k] - self.Ursprung[0]) * position))
            self.ySpiral.append(self.Ursprung[1] + ((self.ySpeichen[k] - self.Ursprung[1]) * position))
        plt.plot(self.xSpiral, self.ySpiral, linewidth = 0.4, color = 'k')

    def makeRoundSpiral(self, windung):
        self.xSpiral = []
        self.ySpiral = []
        ragne = len(self.xSpeichen) * windung
        for i in range (ragne):
            k = i % len(self.xSpeichen)
            position = ((i + 20) / (ragne + 40)) * self.short / Pythagoras((self.xSpeichen[k] - self.Ursprung[0]), (self.ySpeichen[k] - self.Ursprung[1]))
            self.xSpiral.append(self.Ursprung[0] + ((self.xSpeichen[k] - self.Ursprung[0]) * position))
            self.ySpiral.append(self.Ursprung[1] + ((self.ySpeichen[k] - self.Ursprung[1]) * position))
        plt.plot(self.xSpiral, self.ySpiral, linewidth = 0.4, color = 'k')



# berechnet den Schnittpunkt von zwei Geraden (ab und cd)
def line_intersection(ax, bx, cx, dx, ay, by, cy, dy):
    # was hier passiert ist Magie
    xdiff = (ax - bx, cx - dx)
    ydiff = (ay - by, cy - dy)

    def det(a, b):
        return a[0] * b[1] - a[1] * b[0]

    div = det(xdiff, ydiff)
    if div == 0:
        raise Exception('lines do not intersect')

    d = ((ax * by - ay * bx), (cx * dy - cy * dx))
    x = det(d, xdiff) / div
    y = det(d, ydiff) / div
    return x, y    

def Pythagoras(a,b):
    return pow(pow(a, 2) + pow(b, 2), 0.5)

def Average(liste):
    avg = 0
    for i in range (len(liste)):
        avg += liste[i] / len(liste)
    return avg

newNet = Netz(20,3,0,0,1)

plt.show()